#! /usr/bin/python
# Image3D V1.1.1
'''
--------------------------------------------------------------------------------
Image3D Stereo Image Creation

Copyright (C) 2015 Chris Rogers

GNU GENERAL PUBLIC LICENSE GPLv3

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have a copy of the GNU General Public License
    in /usr/share/common-licenses/GPL-3.  If not, see <http://www.gnu.org/licenses/>.

--------------------------------------------------------------------------------
Dependencies

Hugin       >= 2013.0.0 (for align-image-stack)
python-wand >= 0.3.5
python      >= 2.7
ImageMagick >= 6.7

--------------------------------------------------------------------------------
Useful - symbolic link ln -fs "/usr/share/image3d/image3d.py" "/usr/bin/image3d"
      Needs removing after removing package.

Files:
    help, preferences, icon
    program, bash script
    html or py or c file
--------------------------------------------------------------------------------
V1.1 changes from V1.0
    Addition of image display
V1.1.1 changes from V1.1
When changing file, directory or view, change listind to -1 to make sure it starts on first image.
'''

import sys, os, time
from wand.image import Image
from PyQt4 import QtCore, QtGui, uic

#========== main program
# create global variables and set default values
imagesout = 0
listnew = []
listdir = []
listfile = []
listind = -1
progdir = '/usr/share/image3d/'
homedir = os.getenv('HOME') + '/'             		# get homedir directory
workdir = homedir                                       # set current directory
os.chdir(workdir)                               	# start in workdir directory
ext = ''
preffile = 'image3d.dat'
helpfile = 'image3d.html'
bashfile = 'image3dpair.sh'
uifile = 'image3d.ui'
message = ''                                       	# warning message for first use
# create hidden directory for intermediate files
hidedir = homedir + '.image3d/'
result = 0
if not os.path.isdir(hidedir):
    result = os.system('mkdir '+ hidedir)
    if result != 0:
        sys.exit(result)

print 'COPY PREFERENCES FILE'
# copy preferences file
result = 0
if os.path.isfile(hidedir+preffile):
    message = ''
else:
    message = '''This program has the potential to create a very large number of images.
Be careful not to run it in a directory with a large number of image files,
such as a pictures directory with many files with names like DSCN0001.JPG.

Input files must be upright, if they are not, rotate them first.
'''
    result = 'cp ' + progdir + preffile + ' ' + hidedir + preffile
    if result != 0:
        sys.exit(result)

print 'LOAD PREFERENCES'
# create preferences array
prefdata = []
for i in xrange(6):
    prefdata.append([])
    for j in xrange(2):
        prefdata[i].append('')

# load preferences file
with open (hidedir + preffile, 'r') as infile:
    for i in range(0, 6):
        line = infile.readline() ; fields = line.split() ; prefdata[i][0] = fields[0] ; prefdata [i][1] = fields[1]

# use preferences data
if prefdata[0][0] == 'image3d' and prefdata[0][1][1:4] == '1.1':
    if prefdata[1][0] == 'directory' and os.path.exists(prefdata[1][1]):
        workdir = prefdata[1][1]
    else:
        workdir = homedir

    if prefdata[2][0] == 'filename' and os.path.isfile(prefdata[2][1]):
        filename = prefdata[2][1]
    else:
        filename = '{none}'

    if prefdata[3][0] == 'format' and prefdata[3][1] in ['Anaglyph', 'Side-by-Side', 'Crossover']:
        format = prefdata[3][1]
    else:
        format = 'Anaglyph'

    if prefdata[4][0] == 'style' and prefdata[4][1] in ['Level','Popout']:
        style = prefdata[4][1]
    else:
        style = 'Level'

    if prefdata[5][0] == 'view' and prefdata[5][1] in ['New','File','Directory']:
        view = prefdata[5][1]
    else:
        view = 'New'
else:
    # if the file is in the wrong format, use standard values
    workdir = homedir
    filename = '{none}'
    format = 'Anaglyph'
    style = 'Level'
    view = 'New'

formatlet = format[0:1] ; stylelet = style[0:1]
if style == 'Level':
    stylecode = 'A'
else:
    stylecode = 'P'

print 'DEFINITIONS'

#================================================================================
# make file lists
def findfiles(self):
    global listdir, listfile

    listfile = []
    for newfilename in os.listdir(workdir):
        if (newfilename[0:len(filename) + 2] == '3D'+filename and
                (newfilename.endswith('.jpg') or newfilename.endswith('.JPG')
                or newfilename.endswith('.tiff') or newfilename.endswith('.TIFF')
                or newfilename.endswith('.tif') or newfilename.endswith('.TIF')
                or newfilename.endswith('.png') or newfilename.endswith('.PNG'))):
            listfile.append(newfilename)
    listfile.sort()

    listdir = []
    for newfilename in os.listdir(workdir):
        if (newfilename[0:2] == '3D' and (newfilename.endswith('.jpg') or newfilename.endswith('.JPG') or newfilename.endswith('.tiff') or newfilename.endswith('.TIFF') or newfilename.endswith('.tif') or newfilename.endswith('.TIF') or newfilename.endswith('.png') or newfilename.endswith('.PNG'))):
            listdir.append(newfilename)
    listdir.sort()

#================================================================================
def show(self):
    global listind
    locallist = []
    if view == 'New':
        locallist = listnew
    elif view == 'File':
        locallist = listfile
    else:
        locallist = listdir
    if len(locallist) > 0:
        if listind < 0:
            listind = 0
        elif listind > len(locallist) -1:
            listind = len(locallist) -1;
        if os.path.isfile(workdir+locallist[listind]):
            myPixmap = QtGui.QPixmap(workdir+locallist[listind])
            self.label.setScaledContents(True)
            self.label.setPixmap(myPixmap)
            self.textEdit.setText(locallist[listind])
        else:
            self.textEdit.setText('The image '+locallist[listind]+' has not been created yet.')
    else:
        self.textEdit.setText('There are no suitable images yet.')
        self.label.clear()

#===============================actions=================================================
def process_pair(self, workdir, filename, ext, leftn, rightn):
    global imagesout, listtodo, listnew
    # dollar*, img, imageL, imageR are all local variables
    dollar1left = workdir+filename+str(leftn)+ext ; dollar2right = workdir+filename+str(rightn)+ext
    dollar3inter = hidedir+filename+str(leftn)+'+'+str(rightn)+formatlet+stylelet
    dollar4out = workdir+'3D'+filename+str(leftn)+'+'+str(rightn)+formatlet+stylelet+ext
    dollar5format = formatlet                     ; dollar6style = stylecode
    dollar7report = workdir+'3D'+filename+str(leftn)+'+'+str(rightn)+formatlet+stylelet+'.txt'

    with Image(filename = dollar1left) as img:
        imageLF = img.format
        imageLS = img.size
    with Image(filename = dollar2right) as img:
        imageRF = img.format
        imageRS = img.size
    if imageLF == imageRF and imageLS == imageRS:
        result = progdir + bashfile + ' "' + dollar1left +'" "'+ dollar2right +'" "'+ dollar3inter +'" "'+ dollar4out +'" '+  dollar5format +' '+  dollar6style + ' "' + dollar7report + '" &'
        result = os.system(result)
        if result != 0:
            self.textEdit.setText('The Bash program has failed to start')
        self.textEdit.append('Creating 3D'+filename+str(leftn)+'+'+str(rightn)+formatlet+stylelet+ext)
        listnew.append(dollar4out)
        imagesout = imagesout + 1
    else:
        if imageLF != imageRF:
            self.textEdit.append('The files ' + filename+str(leftn)+ext + ' and ' + filename+str(rightn)+ext + ' can not be used as they are different types of file.')
        if imageLS != imageRS:
            self.textEdit.append('The files ' + filename+str(leftn)+ext + ' and ' + filename+str(rightn)+ext + ' can not be used as they are different sizes.')

#================================================================================
def process_set(self, workdir, filename, ext):
    global imagesout, listtodo, listnew
    # find out how many images there are in this set
    imagesin = 0 # count of number of input files
    for num in range (0, 9):
        if os.path.isfile(workdir + filename + str(num) + ext):
            imagesin = imagesin + 1

    imagesout = 0
    if imagesin == 0:
        self.textEdit.append('There are no suitable images in the image set ' + filename + '*' + ext + '.')
    elif imagesin == 1:
        self.textEdit.append('There is only one file in the image set ' + filename + '*' + ext + ', at least two files are needed.')
    else:
        self.textEdit.append('Checking the image set ' + filename + '*' + ext + ', which has ' + str(imagesin) + ' images.')
        imagesout = 0 # count of number of 3D images which should be created
        # loop through all image pairs
        leftn = 0 ; rightn = 1
        while leftn < imagesin + 1:
            rightn = leftn + 1
            while rightn < imagesin + 1:
                if os.path.isfile(workdir + '3D' + filename + str(leftn) + '+' + str(rightn) + formatlet + stylelet + ext):
                    self.textEdit.append('3D' + filename + str(leftn) + '+' + str(rightn) + formatlet + stylelet + ext + ' already exists.')
                elif not (os.path.isfile(workdir + filename + str(leftn) + ext) and os.path.isfile(workdir + filename + str(rightn) + ext)):
                    dummy = 0 # There is a gap in the sequence of file numbers
                else:
                    process_pair(self, workdir, filename, ext, leftn, rightn)
                rightn = rightn + 1
            leftn = leftn + 1

    if imagesout < 1:
        self.textEdit.append('No 3D images will created for this set.\n')
    elif imagesout == 1:
        self.textEdit.append('One 3D image should be created for this set, please wait...\n')
    else:
        self.textEdit.append(str(imagesout) + ' 3D images should created for this set, please wait...\n')

print 'LOAD THE GUI'
#========== GUI
# Load the UI
form_class = uic.loadUiType(progdir + uifile)[0]

# define UI class and actions
class MyWindowClass(QtGui.QMainWindow, form_class):
    def __init__(self, parent=None):
        QtGui.QMainWindow.__init__(self, parent)
        self.setupUi(self)
        # Bind the event handlers to the radio buttons, no need for Quit as it is handled by GUI
        self.actionDirectory.triggered.connect(self.actionDirectoryTriggered)
        self.actionFile.triggered.connect(self.actionFileTriggered)
        self.actionPrefer.triggered.connect(self.actionPreferTriggered)
        self.actionHelp.triggered.connect(self.actionHelpTriggered)
        self.actionAbout.triggered.connect(self.actionAboutTriggered)

        self.radioFormatAnaglyph.clicked.connect(self.radioFormatAnaglyphClicked)
        self.radioFormatSidebyside.clicked.connect(self.radioFormatSidebysideClicked)
        self.radioFormatCrossover.clicked.connect(self.radioFormatCrossoverClicked)
        self.radioStyleLevel.clicked.connect(self.radioStyleLevelClicked)
        self.radioStylePopout.clicked.connect(self.radioStylePopoutClicked)
        self.radioViewNew.clicked.connect(self.radioViewNewClicked)
        self.radioViewFile.clicked.connect(self.radioViewFileClicked)
        self.radioViewDirectory.clicked.connect(self.radioViewDirectoryClicked)
        self.buttonProcessOne.clicked.connect(self.buttonProcessOneClicked)
        self.buttonProcessAll.clicked.connect(self.buttonProcessAllClicked)
        self.buttonLeft.clicked.connect(self.buttonLeftClicked)
        self.buttonRight.clicked.connect(self.buttonRightClicked)
        self.fileDialog = QtGui.QFileDialog(self)

        # to fix obscure bug
        self.radioFormatAnaglyph.setStyle(QtGui.QStyleFactory.create("Cleanlooks"))
        self.radioFormatSidebyside.setStyle(QtGui.QStyleFactory.create("Cleanlooks"))
        self.radioFormatCrossover.setStyle(QtGui.QStyleFactory.create("Cleanlooks"))
        self.radioStyleLevel.setStyle(QtGui.QStyleFactory.create("Cleanlooks"))
        self.radioStylePopout.setStyle(QtGui.QStyleFactory.create("Cleanlooks"))
        self.radioViewNew.setStyle(QtGui.QStyleFactory.create("Cleanlooks"))
        self.radioViewFile.setStyle(QtGui.QStyleFactory.create("Cleanlooks"))
        self.radioViewDirectory.setStyle(QtGui.QStyleFactory.create("Cleanlooks"))

        print 'SET PREFERENCES'
        if format == 'Anaglyph':
            self.radioFormatAnaglyph.setChecked(True)
        elif format == 'Side-by-Side':
            self.radioFormatSidebyside.setChecked(True)
        else:
            self.radioFormatCrossover.setChecked(True)

        if style == 'Level':
            self.radioStyleLevel.setChecked(True)
        else:
            self.radioStylePopout.setChecked(True)

        if view == 'New':
            self.radioViewNew.setChecked(True)
        elif view == 'File':
            self.radioViewFile.setChecked(True)
        else:
            self.radioViewDirectory.setChecked(True)

        self.statusBar().showMessage(self.tr('Directory: ' + workdir + '          Image Set: ' + filename + '*' + ext))
        self.setWindowTitle('Image3D')

        # on first use show warning about large numbers of files
        self.textEdit.append(message)

        print 'MAKE FILE LIST'
        findfiles(self)

    print 'BUTTON DEFINITIONS'
    # settings
    def actionPreferTriggered(self):
        # local file
        self.statusBar().showMessage(self.tr('Directory: ' + workdir + '          Image Set: ' + filename + '*' + ext))
        result = QtGui.QMessageBox.question(self, 'Preferences', 'Save your current setting as the default settings?', QtGui.QMessageBox.No, QtGui.QMessageBox.Yes)
        if result == 16384: # Yes
            with open (hidedir + preffile, 'w') as file:
                file.write('image3d V1.1\n')
                file.write('directory '+workdir+'\n')
                file.write('filename '+filename+'\n')
                file.write('format '+format+'\n')
                file.write('style '+style+'\n')
                file.write('view '+view+'\n')

    # button actions (No need for Quit button definition as it is handled by GUI)
    def actionDirectoryTriggered(self):
        global workdir, listdir, listind
        self.statusBar().showMessage(self.tr('Directory: ' + workdir + '          Image Set: ' + filename + '*' + ext))
        newdir = QtGui.QFileDialog.getExistingDirectory(self, 'Select the directory where the images are', directory = homedir) +'/'
        if newdir[0:len(homedir)] == homedir:
            workdir = newdir; os.chdir(workdir)
            findfiles(self); listind = -1
        else:
            result = QtGui.QMessageBox.question(self, 'Warning', 'Directory not changed', QtGui.QMessageBox.Close)
        self.statusBar().showMessage(self.tr('Directory: ' + workdir + '          Image Set: ' + filename + '*' + ext))

    # menu bar actions
    def actionFileTriggered(self):
        global workdir, filename, ext, listdir, listind
        # local newfilename, newdir
        self.statusBar().showMessage(self.tr('Directory: ' + workdir + '          Image Set: ' + filename + '*' + ext))
        newfilename = QtGui.QFileDialog.getOpenFileName(self, 'Select any file from the set to process.', directory = workdir, filter = '*.jpg *.JPG *.tiff *.TIFF *.tif *.TIF *.png *.PNG')
        newfilename = str(newfilename) # filename is not string type
        if newfilename[0:len(homedir)] == homedir:
            newdir, newfilename = os.path.split(newfilename)
            if newfilename[0:2] != '3D':
                workdir = newdir + '/'; os.chdir(workdir)
                filename, ext = os.path.splitext(newfilename) ; filename = filename[:-1]
                findfiles(self); listind = -1
            else:
                result = QtGui.QMessageBox.question(self, 'Warning', 'No image selected', QtGui.QMessageBox.Close)
        else:
            result = QtGui.QMessageBox.question(self, 'Warning', 'No image selected', QtGui.QMessageBox.Close)
        self.statusBar().showMessage(self.tr('Directory: ' + workdir + '          Image Set: ' + filename + '*' + ext))

    def actionHelpTriggered(self):
        self.statusBar().showMessage(self.tr('Directory: ' + workdir + '          Image Set: ' + filename + '*' + ext))
        with open(progdir + helpfile, 'r') as infile:
            result = infile.read()
        self.helptext = QtGui.QTextEdit()  # self would tie window to main window, but window can't be closed without an OK button
        self.helptext.setHtml(result)
        self.helptext.setReadOnly(-1)
        self.helptext.resize(800, 500)
        self.helptext.show()

    def actionAboutTriggered(self):
        self.statusBar().showMessage(self.tr('Directory: ' + workdir + '          Image Set: ' + filename + '*' + ext))
        result = QtGui.QMessageBox.about(self, 'About this program','''

        Image3D Stereo Image Creation V1.0

        Copyright (C) 2015 Chris Rogers

        GNU GENERAL PUBLIC LICENSE

        The clever stuff is done by:
            Hugin - Image alignment
            http://hugin.sourceforge.net/

            ImageMagick - 3D image composition
            http://www.imagemagick.org/

        ''')

    # format actions
    def radioFormatAnaglyphClicked(self):
        global format, formatlet
        format = 'Anaglyph' ; formatlet = 'A'

    def radioFormatSidebysideClicked(self):
        global format, formatlet
        format = 'Side-by-Side' ; formatlet = 'S'

    def radioFormatCrossoverClicked(self):
        global format, formatlet
        format = 'Crossover' ; formatlet = 'C'

    # style actions
    def radioStyleLevelClicked(self):
        global style, stylelet, stylecode
        style = 'Level' ; stylelet = 'L' ; stylecode = 'A'

    def radioStylePopoutClicked(self):
        global style, stylelet, stylecode
        style = 'Popout' ; stylelet = 'P' ; stylecode = 'P'

    # process actions
    # only one image pair
    def buttonProcessOneClicked(self):
        global imagesout, listnew
        listnew = []
        self.textEdit.clear()
        if filename != '{none}':
            process_set(self, workdir, filename, ext)
        else:
            result = QtGui.QMessageBox.question(self, 'Warning', 'A file must be selected first.', QtGui.QMessageBox.Close)
        for newfilename in listnew:
            if newfilename not in listdir:
                listdir.append(newfilename)

    # loop through all image pairs
    def buttonProcessAllClicked(self):
        global imagesout, listnew
        listnew = []
        self.textEdit.clear()
        listtodo = []; filename = ''; ext = ''
        for newfilename in os.listdir(workdir):
            if (newfilename[0:2] != '3D' and (newfilename.endswith('.jpg') or newfilename.endswith('.JPG') or newfilename.endswith('.tiff') or newfilename.endswith('.TIFF') or newfilename.endswith('.tif') or newfilename.endswith('.TIF') or newfilename.endswith('.png') or newfilename.endswith('.PNG'))):
                filename, ext = os.path.splitext(newfilename) ; filename = filename[:-1]
                if filename not in listtodo:
                    listtodo.append(filename)
                    process_set(self, workdir, filename, ext)
        if listtodo == []:
            self.textEdit.append('There are no suitable images.')
        for newfilename in listnew:
            if newfilename not in listdir:
                listdir.append(newfilename)

    def radioViewNewClicked(self):
        global view
        view = 'New'
        listind = -1
        self.label.clear()

    def radioViewFileClicked(self):
        global view
        view = 'File'
        listind = -1
        self.label.clear()

    def radioViewDirectoryClicked(self):
        global view
        view = 'Directory'
        listind = -1
        self.label.clear()

    def buttonLeftClicked(self):
        global listind
        listind = listind - 1
        show(self)

    def buttonRightClicked(self):
        global listind
        listind = listind + 1
        show(self)

#========== start GUI
print 'START GUI'
app = QtGui.QApplication(sys.argv)
myWindow = MyWindowClass(None)
myWindow.show()
app.exec_()
print 'END PROGRAM'
#========== end

